from __future__ import annotations

import json
from typing import Any

from .const import Feature, OperationMode


class ElectraAirConditioner(object):
    def __init__(self, data: dict[str, str]) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.regdate: str = data["regdate"]
        self.model = data["model"]
        self.mac: str = data["mac"]
        self.serial_number: str = data["sn"]
        self.manufactor: str = data["manufactor"]
        self.type: str = data["deviceTypeName"]
        self.status: str = data["status"]
        self.token: str = data["deviceToken"]
        self._time_delta: int = 0
        self.features: list[int] = []

        self.current_mode: str | None = None
        self.collected_measure: int | None = None
        self._oper_data: dict[str, str] = {}

    def is_disconnected(self, thresh_sec: int = 60) -> bool:
        return self._time_delta > thresh_sec

    def update_features(self) -> None:
        if "VSWING" in self._oper_data:
            self.features.append(Feature.V_SWING)
        if "HSWING" in self._oper_data:
            self.features.append(Feature.H_SWING)

    def get_sensor_temperature(self) -> int | None:
        return self.collected_measure

    def get_mode(self) -> str:
        return self._oper_data["AC_MODE"]

    def set_mode(self, mode: str) -> None:
        if mode in [
            OperationMode.MODE_AUTO,
            OperationMode.MODE_COOL,
            OperationMode.MODE_DRY,
            OperationMode.MODE_FAN,
            OperationMode.MODE_HEAT,
        ]:
            if mode != self._oper_data["AC_MODE"]:
                self._oper_data["AC_MODE"] = mode

    def set_horizontal_swing(self, enable: bool) -> None:
        if "HSWING" in self._oper_data:
            self._oper_data["HSWING"] = OperationMode.ON if enable else OperationMode.OFF

    def set_vertical_swing(self, enable: bool) -> None:
        if "VSWING" in self._oper_data:
            self._oper_data["VSWING"] = OperationMode.ON if enable else OperationMode.OFF

    def is_vertical_swing(self) -> bool:
        if "VSWING" in self._oper_data:
            return self._oper_data["VSWING"] == OperationMode.ON
        return False

    def is_horizontal_swing(self) -> bool:
        if "HSWING" in self._oper_data:
            return self._oper_data["HSWING"] == OperationMode.ON
        return False

    def is_on(self) -> bool:
        if "TURN_ON_OFF" in self._oper_data:
            return self._oper_data["TURN_ON_OFF"] == OperationMode.ON
        else:
            return self._oper_data["AC_MODE"] != OperationMode.STANDBY

    def turn_on(self) -> None:
        if not self.is_on():
            if "TURN_ON_OFF" in self._oper_data:
                self._oper_data["TURN_ON_OFF"] = OperationMode.ON

    def turn_off(self) -> None:
        if self.is_on():
            if "TURN_ON_OFF" in self._oper_data:
                self._oper_data["TURN_ON_OFF"] = OperationMode.OFF
            else:
                self._oper_data["AC_MODE"] = OperationMode.STANDBY

    def get_temperature(self) -> int:
        return int(self._oper_data["SPT"])

    def set_temperature(self, val: int) -> None:
        if self.get_temperature() != val:
            self._oper_data["SPT"] = str(val)

    def get_fan_speed(self) -> str:
        return self._oper_data["FANSPD"]

    def set_fan_speed(self, speed: str) -> None:
        if speed in [
            OperationMode.FAN_SPEED_AUTO,
            OperationMode.FAN_SPEED_HIGH,
            OperationMode.FAN_SPEED_MED,
            OperationMode.FAN_SPEED_LOW,
        ]:
            if speed != self._oper_data["FANSPD"]:
                self._oper_data["FANSPD"] = speed

    def set_turbo_mode(self, enable: bool) -> None:
        self._oper_data["TURBO"] = OperationMode.ON if enable else OperationMode.OFF

    def get_turbo_mode(self) -> bool:
        return self._oper_data["TURBO"] == OperationMode.ON

    def set_shabat_mode(self, enable: bool) -> None:
        self._oper_data["SHABAT"] = OperationMode.ON if enable else OperationMode.OFF

    def get_shabat_mode(self) -> bool:
        return self._oper_data["SHABAT"] == OperationMode.ON

    def update_operation_states(self, data: dict[str, Any]) -> None:
        self._oper_data = json.loads(data["commandJson"]["OPER"])["OPER"]
        self._time_delta = data["timeDelta"]
        measurments = json.loads(data["commandJson"]["DIAG_L2"])["DIAG_L2"]
        if "I_RAT" in measurments:
            self.collected_measure = int(measurments["I_RAT"])
        if "I_CALC_AT" in measurments:
            self.collected_measure = int(measurments["I_CALC_AT"])

        self.current_mode = measurments["O_ODU_MODE"]

    def get_operation_state(self) -> str:
        if "AC_STSRC" in self._oper_data:
            self._oper_data["AC_STSRC"] = "WI-FI"

        return json.dumps({"OPER": self._oper_data})



from dataclasses import dataclass
from datetime import timedelta

STATUS_SUCCESS = 0
DELAY_BETWEEM_SID_REQUESTS = int(timedelta(minutes=5).total_seconds())
SID_EXPIRATION = int(timedelta(minutes=15).total_seconds())


@dataclass
class Attributes:
    INTRUDER_LOCKOUT = "Intruder lockout"
    DESC = "res_desc"
    DATA = "data"
    TOKEN = "token"
    SID = "sid"
    RES = "res"
    STATUS = "status"
    DEVICES = "devices"



from math import floor, pow
from random import randint


def generate_imei() -> str:
    minimum = int(pow(10, 7))
    maximum = int(pow(10, 8) - 1)
    return f"2b950000{str(floor(randint(minimum, maximum)))}"



from dataclasses import dataclass

MAX_TEMP = 30
MIN_TEMP = 17


@dataclass
class OperationMode:
    MODE_COOL = "COOL"
    MODE_HEAT = "HEAT"
    MODE_AUTO = "AUTO"
    MODE_DRY = "DRY"
    MODE_FAN = "FAN"
    FAN_SPEED_LOW = "LOW"
    FAN_SPEED_MED = "MED"
    FAN_SPEED_HIGH = "HIGH"
    FAN_SPEED_AUTO = "AUTO"
    ON = "ON"
    OFF = "OFF"
    STANDBY = "STBY"


@dataclass
class Feature:
    V_SWING = 0
    H_SWING = 1



oper_state = {
    "id": 99,
    "status": 0,
    "desc": None,
    "data": {
        "timeDelta": 15,
        "commandJson": {
            "OPER": '{"OPER":{"TURN_ON_OFF":"OFF","AC_MODE":"COOL","SPT":"24","FANSPD":"AUTO","VSWING":"OFF","SLEEP":"OFF","HSWING":"OFF","CLEAR_FILT":"OFF","IDU_PN":"","IFEEL":"OFF","MSGTYPE":"OPER","OP_VAL_ERR":"OK","SHABAT":"OFF","TIMER":"OFF","TURBO":"OFF"}}',
            "DIAG_L2": '{"DIAG_L2":{"I_RAT":"24","O_ODU_MODE":"COOL","IDU_FAN":"AUTO","IDU_PN":"","MSGTYPE":"DIAG_L2","O_OAT":""}}',
        },
        "res": 0,
        "res_desc": None,
    },
}

resp = {
    "id": 99,
    "status": 0,
    "desc": None,
    "data": {
        "devices": [
            {
                "providerName": None,
                "deviceTypeName": "A/C",
                "manufactor": "Midea",
                "photoId": None,
                "permissions": 15,
                "deviceTypeId": 1,
                "name": "סלון",
                "status": 1,
                "providerid": 1,
                "latitude": None,
                "longitude": None,
                "location": None,
                "sn": "XXXXXX",
                "mac": "XXXXX",
                "model": "XXXXX",
                "hwVersion": None,
                "fmVersion": None,
                "userId": 99999,
                "manufactorId": 2,
                "iconId": "1",
                "hasImage": False,
                "deviceToken": "XXXXX",
                "mqttId": "XXXXX",
                "enableEvents": True,
                "isActivated": False,
                "logLevel": None,
                "lastIntervalActivity": None,
                "PowerOnID": None,
                "IsDebugMode": False,
                "regdate": "2021-03-31T21:12:39",
                "id": 9999,
            }
        ],
        "res": 0,
        "res_desc": None,
    },
}



from __future__ import annotations

from asyncio import TimeoutError, create_task
from datetime import datetime
from json import JSONDecodeError
from logging import getLogger
from typing import Any, List

from aiohttp import ClientError, ClientSession

from electrasmart.device import ElectraAirConditioner

from .const import (DELAY_BETWEEM_SID_REQUESTS, SID_EXPIRATION, STATUS_SUCCESS,
                    Attributes)

logger = getLogger(__name__)


class ElectraApiError(Exception):
    pass


class ElectraAPI(object):
    def __init__(
        self,
        websession: ClientSession,
        imei: str | None = None,
        token: str | None = None,
    ) -> None:
        self._base_url = "https://app.ecpiot.co.il/mobile/mobilecommand"
        self._sid = None
        self._imei = imei
        self._token = token
        self._sid_expiration = 0
        self._last_sid_request_ts = 0
        self._session = websession
        self._phone_number = None
        self._devices: list[ElectraAirConditioner] = []

        logger.debug("Initialized Electra API object")

    @property
    def devices(self) -> list[ElectraAirConditioner]:
        return self._devices

    async def _send_request(self, payload: dict[str, Any]) -> dict[str, Any]:
        try:
            resp = await self._session.post(
                url=self._base_url,
                json=payload,
                headers={"user-agent": "Electra Client"},
            )
            json_resp: dict[str, Any] = await resp.json(content_type=None)
        except TimeoutError as ex:
            raise ElectraApiError(f"Failed to communicate with Electra API due to time out: ({str(ex)})")
        except ClientError as ex:
            raise ElectraApiError(f"Failed to communicate with Electra API due to ClientError: ({str(ex)})")
        except JSONDecodeError as ex:
            raise ElectraApiError(f"Recieved invalid response from Electra API: {str(ex)}")

        return json_resp

    async def generate_new_token(self, phone_number: str, imei: str) -> dict[str, Any]:
        payload = {
            "pvdid": 1,
            "id": 99,
            "cmd": "SEND_OTP",
            "data": {"imei": imei, "phone": phone_number},
        }

        return await self._send_request(payload=payload)

    async def validate_one_time_password(self, otp: str, imei: str, phone_number: str) -> dict[str, Any]:
        payload = {
            "pvdid": 1,
            "id": 99,
            "cmd": "CHECK_OTP",
            "data": {
                "imei": imei,
                "phone": phone_number,
                "code": otp,
                "os": "android",
                "osver": "M4B30Z",
            },
        }

        return await self._send_request(payload=payload)

    def _sid_expired(self) -> bool:
        current_time = int(datetime.now().timestamp())
        refresh_in = self._sid_expiration - current_time
        if refresh_in > 0:
            logger.debug("Should refresh in %s minutes", round(refresh_in / 60))

        if current_time < self._sid_expiration:
            return False
        else:
            self._sid = None
            return True

    async def _get_sid(self, force: bool = False) -> None:
        current_ts = int(datetime.now().timestamp())
        if not force and not self._sid_expired():
            logger.debug("Found valid sid (%s) in cache, using it", self._sid)
            return

        if self._last_sid_request_ts and current_ts < (self._last_sid_request_ts + DELAY_BETWEEM_SID_REQUESTS):
            logger.debug(
                'Session ID was requested less than 5 minutes ago! waiting in order to prevent "intruder lockdown"...'
            )
            return

        payload = {
            "pvdid": 1,
            "id": 99,
            "cmd": "VALIDATE_TOKEN",
            "data": {
                "imei": self._imei,
                "token": self._token,
                "os": "android",
                "osver": "M4B30Z",
            },
        }

        resp = await self._send_request(payload=payload)

        if resp is None:
            raise ElectraApiError("Failed to retrieve sid")
        else:
            if not resp[Attributes.DATA][Attributes.SID]:
                raise ElectraApiError(
                    "Failed to retrieve SID due to %s",
                    resp[Attributes.DATA][Attributes.DESC],
                )

            else:
                self._sid = resp[Attributes.DATA][Attributes.SID]
                self._sid_expiration = current_ts + SID_EXPIRATION
                self._last_sid_request_ts = current_ts
                logger.debug("Successfully acquired sid: %s", self._sid)

    async def fetch_devices(self) -> None:
        fetch_state_tasks = []
        logger.debug("About to Get Electra AC devices")
        await self._get_sid()

        payload = {"pvdid": 1, "id": 99, "cmd": "GET_DEVICES", "sid": self._sid}

        ac_list: List[ElectraAirConditioner] = []
        resp = await self._send_request(payload=payload)
        if resp[Attributes.STATUS] == STATUS_SUCCESS:
            for ac in resp[Attributes.DATA][Attributes.DEVICES]:
                if ac["deviceTypeName"] == "A/C":
                    electra_ac: ElectraAirConditioner = ElectraAirConditioner(ac)
                    logger.debug("Discovered A/C device %s", electra_ac.name)
                    ac_list.append(electra_ac)
                    fetch_state_tasks.append(create_task(self.get_last_telemtry(electra_ac)))
                else:
                    logger.debug("Discovered non AC device %s", ac)

            for task in fetch_state_tasks:
                await task

            for ac in ac_list:
                ac.update_features()

            self._devices = ac_list

        else:
            raise ElectraApiError("Failed to fetch devices %s", resp)

    async def get_last_telemtry(self, ac: ElectraAirConditioner) -> None:
        logger.debug("Getting AC %s state", ac.name)

        await self._get_sid()

        payload = {
            "pvdid": 1,
            "id": 99,
            "cmd": "GET_LAST_TELEMETRY",
            "sid": self._sid,
            "data": {"id": ac.id, "commandName": "OPER,DIAG_L2"},
        }

        resp = await self._send_request(payload=payload)
        if resp[Attributes.STATUS] != STATUS_SUCCESS:
            raise ElectraApiError(f"Failed to get AC operation state: {resp}")
        else:
            ac.update_operation_states(resp[Attributes.DATA])

    async def set_state(self, device: ElectraAirConditioner) -> dict[str, Any]:
        json_command = device.get_operation_state()
        await self._get_sid()

        payload = {
            "pvdid": 1,
            "id": 99,
            "cmd": "SEND_COMMAND",
            "sid": self._sid,
            "data": {"id": device.id, "commandJson": json_command},
        }

        return await self._send_request(payload=payload)
